/*
=======================================================================

APC Test driver

By Enrico Martignetti


=======================================================================
*/


//----------------------------------------------------------------------
//
// Includes
// --------


extern "C" {
#include <NTDDK.h>

#include "ApcDefs.h"
#include "KernelHooks.h"
#include "MiscKDefs.h"
}

#include "CircBuffer.h"


// Defs shared with ring 3 components.
//

#include "DrvR3.h"

// Definition of the internal device name and symbolic link. They are 
// derived from DRV_DEVICE_NAME which is the public name defined
// into DrvR3.h and used by client applications.

#define DRV_INT_NAME    L"\\Device\\"DRV_DEVICE_NAME
#define DRV_SYM_LINK    L"\\??\\"DRV_DEVICE_NAME


//----------------------------------------------------------------------
//
// Defines
// -------

// Offset of the APC interrupt flag inside _KPCR
#define APC_INT_FLAG_OFFSET 0x96

// Offset of ApcState inside _KTHREAD
#define APC_STATE_OFFSET	0x38

// Offset of KernelApcPending inside _KAPC_STATE
#define KERNEL_APC_PENDING_OFFSET	0x15

// Offset of the IRQL field inside the _KPCR
#define IRQL_OFFSET	0x24

// Offset of KeWaitForSingleObject inside ntoskrnl.exe
//
#define OFF_KEWAITFORSINGLEOBJECT 0x547c9

// Offset of SwapContext inside ntoskrnl.exe
//
#define OFF_SWAPCONTEXT			0x690a0

// Offset of KiDeliverApc inside ntoskrnl.exe
//
#define OFF_KIDELIVERAPC		0x7a08c

// Offset of UserApcPending inside _KAPC_STATE
//
#define USER_APC_PENDING_OFFSET 0x16
//----------------------------------------------------------------------
//
// Data Types
// ----------


typedef struct _DEVICE_EXTENSION {
	PDEVICE_OBJECT pDevice;

    KAPC	Apc;

	KEVENT Event;


} DEVICE_EXTENSION, *PDEVICE_EXTENSION;

typedef enum {
	SwapContextRecord,
	KiDeliverApcRecord,
	KernelRoutineRecord,
	MessageRecord
} RECORD_TYPE;

typedef struct _THREAD_INFO {
	PKTHREAD	Thread;
	BOOLEAN		KernelApcPending;
	BOOLEAN		KApcListEmpty;
	BOOLEAN		UserApcPending;
	BOOLEAN		UApcListEmpty;
} THREAD_INFO, *PTHREAD_INFO;

typedef struct _KI_DELIVER_APC_RECORD {
	RECORD_TYPE	RecordType;
	KIRQL		Irql;
	PCHAR		Caller;
	PCHAR		TrapFrame;
	PVOID		Reserved;
	INT			PreviousMode;
	THREAD_INFO	Thread;
} KI_DELIVER_APC_RECORD, *PKI_DELIVER_APC_RECORD;

typedef struct _SWAP_CONTEXT_RECORD {
	RECORD_TYPE	RecordType;
	KIRQL		Irql;
	THREAD_INFO	CurrThr;
	THREAD_INFO	NewThr;
	BOOLEAN		ApcInt;
} SWAP_CONTEXT_RECORD, *PSWAP_CONTEXT_RECORD;

typedef struct _KERNEL_ROUTINE_RECORD {
	RECORD_TYPE	RecordType;
	THREAD_INFO	Thread;
} KERNEL_ROUTINE_RECORD, *PKERNEL_ROUTINE_RECORD;

typedef struct _MESSAGE_RECORD {
	RECORD_TYPE	RecordType;
	INT			MessageIndex;
} MESSAGE_RECORD, *PMESSAGE_RECORD;

typedef union _TRACE_RECORD {
	RECORD_TYPE				RecordType;
	SWAP_CONTEXT_RECORD		SwapContextRecord;
	KI_DELIVER_APC_RECORD	DeliverApcRecord;
	KERNEL_ROUTINE_RECORD	KernelRoutineRecord;
	MESSAGE_RECORD			MessageRecord;
} TRACE_RECORD, *PTRACE_RECORD;


extern "C" {
__declspec(dllimport) extern void __fastcall HalRequestSoftwareInterrupt(
	unsigned long	par1);
}

BOOLEAN bBufferOverflow;
BOOLEAN bHookTraceOn;
CHAR brkByte;
CIRCULAR_BUFFER BufCtl;
CHAR BufferStore[4 * 1024];
unsigned long InitCtxSwitches;
unsigned long InApcCtxSwitches;
extern "C" {
	PCHAR lpKiDeliverApc = 0;
	PCHAR lpNtOsKrnl = 0;
	PCHAR lpSwapContext = 0;
}
PKTHREAD pkCurrThr;
CHAR InitIntFlg;
PCHAR szTraceMessages[] = {
	"Returned from KeLowerIrql"
};


//----------------------------------------------------------------------
//
// Forward declarations
// --------------------


static VOID DriverUnload (
		IN PDRIVER_OBJECT	pDriverObject	);

static NTSTATUS CreateDevice (
		IN PDRIVER_OBJECT	pDriverObject,
		OUT PDEVICE_EXTENSION*	ppDevExt	);

static VOID DumpTrace();

static BOOLEAN ApcInt(
	PCHAR		pKpcr);

static VOID ApcKernelRoutine (
	IN struct _KAPC *Apc,
	IN OUT PKNORMAL_ROUTINE *NormalRoutine,
	IN OUT PVOID *NormalContext,
	IN OUT PVOID *SystemArgument1,
	IN OUT PVOID *SystemArgument2 );

static VOID ApcKernelRoutine2 (
	IN struct _KAPC *Apc,
	IN OUT PKNORMAL_ROUTINE *NormalRoutine,
	IN OUT PVOID *NormalContext,
	IN OUT PVOID *SystemArgument1,
	IN OUT PVOID *SystemArgument2 );

static VOID ApcRequester(
	 IN PVOID		StartContext);

static NTSTATUS ApcScheduleTest(
	PDEVICE_OBJECT		pDevObj, 
	PIRP				pIrp);

static NTSTATUS ApcSpuriousIntTest(
	PDEVICE_OBJECT		pDevObj, 
	PIRP				pIrp);

static NTSTATUS ApcUserModeTest(
	PDEVICE_OBJECT		pDevObj, 
	PIRP				pIrp);

static NTSTATUS ApcUserModeTes2(
	PDEVICE_OBJECT		pDevObj, 
	PIRP				pIrp);

static NTSTATUS DeleteDevices(
		IN PDRIVER_OBJECT	pDriverObject	);

static NTSTATUS DispatchCreate (
		IN PDEVICE_OBJECT	pDevObj,
		IN PIRP				pIrp			);

static NTSTATUS DispatchClose (
		IN PDEVICE_OBJECT	pDevObj,
		IN PIRP				pIrp			);

static NTSTATUS DispatchDioc (
		IN PDEVICE_OBJECT	pDevObj,
		IN PIRP				pIrp			);

NTSTATUS DispatchRead(
	  IN PDEVICE_OBJECT	pDevObj,
	  IN PIRP			pIrp );

NTSTATUS DispatchWrite(
	  IN PDEVICE_OBJECT	pDevObj,
	  IN PIRP			pIrp );

static BOOLEAN GetKernelApcPending(
		PCHAR		lpCurrThr);

static PCHAR GetKpcr();

static KIRQL GetIrql();

static BOOLEAN KApcListEmpty(
	PKTHREAD		Thread);

static BOOLEAN KernelApcPending(
		PKTHREAD		Thread);

VOID KApcNormalRoutine(
	   PVOID		NormalContext,
	   PVOID		SystemArgument1,
	   PVOID		SystemArgument2);

extern "C" VOID __cdecl KiDeliverApcTrace(
  PCHAR		TrapFrame,
  PVOID		Reserved,
  INT		PreviousMode,
  PCHAR		pCaller,
  PKTHREAD	CurrThr);

static VOID PrintTraceRecord(
	PTRACE_RECORD	pTrcRec);

static VOID RemoveHooks();

static VOID SetHooks();

extern "C" VOID __cdecl SwapContextTrace(
  PKTHREAD		lpCurrThr,
  PKTHREAD		lpNewThr);

static VOID UApcKernelRoutine (
	IN struct _KAPC *Apc,
	IN OUT PKNORMAL_ROUTINE *NormalRoutine,
	IN OUT PVOID *NormalContext,
	IN OUT PVOID *SystemArgument1,
	IN OUT PVOID *SystemArgument2 );

static BOOLEAN UApcListEmpty(
	PKTHREAD	Thread);

static BOOLEAN UserApcPending(
	PKTHREAD	Thread);




//----------------------------------------------------------------------
//
// Functions
// ---------


//++
// Function:	DriverEntry
//
// Description:
//
//      Announces entry points.
//
//		Creates the device object used to send control codes to this
//		module.
//
// Arguments:
//		pDriverObject - Passed from I/O Manager
//		pRegistryPath - UNICODE_STRING pointer to
//						registry info (service key)
//						for this driver
//
// Return value:
//		NTSTATUS signaling success or failure
//--
extern "C" NTSTATUS DriverEntry (
			IN PDRIVER_OBJECT pDriverObject,
			IN PUNICODE_STRING pRegistryPath	) {
	NTSTATUS status;
	PDEVICE_EXTENSION	pDevExt;




	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - APC test driver, compiled %s %s\n",__DATE__, __TIME__);

	KPCR TestPcr;
	SIZE_T offset;
	offset = ((PCHAR *) &TestPcr.Irql) - ((PCHAR *) &TestPcr);
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - offset of Irql = %#x", offset);



	// Announce other driver entry points
	pDriverObject->DriverUnload = DriverUnload;
	pDriverObject->MajorFunction[IRP_MJ_CREATE] =
				DispatchCreate;
	pDriverObject->MajorFunction[IRP_MJ_CLOSE] =
				DispatchClose;
	pDriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] =
				DispatchDioc;
	pDriverObject->MajorFunction[IRP_MJ_READ] = 
				DispatchRead;
	pDriverObject->MajorFunction[IRP_MJ_WRITE] = 
				DispatchWrite;
	
	// Create the logical device
	status =
		CreateDevice(pDriverObject, &pDevExt);
	if (!NT_SUCCESS(status))
		return status;



	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - driver succesfully loaded.");

	return STATUS_SUCCESS;
}


//++
// Function:	DriverUnload
//
// Description:
//
//      Realases any resource stored into the device extension,
//      then realases the device object.
//
// Arguments:
//		pDriverObject - Passed from I/O Manager
//
// Return value:
//		None
//--

VOID DriverUnload (
		IN PDRIVER_OBJECT	pDriverObject	) {

    PDEVICE_EXTENSION       lpDevExt;

    // Obtain a pointer to the device extension
    lpDevExt = (PDEVICE_EXTENSION)
        pDriverObject->DeviceObject->DeviceExtension;

    // Release any resources stored into the device extension here.

	// Delete the logical device.
	DeleteDevices(pDriverObject);

	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - driver unloaded.");


	return;
}


static BOOLEAN ApcInt(
	PCHAR		pKpcr)
{

	return *(pKpcr + APC_INT_FLAG_OFFSET);
}


//++
// Function:	
//
// Description:
//
// Arguments:
//
// Return value:
//
//--
static VOID ApcKernelRoutine (
	IN struct _KAPC *Apc,
	IN OUT PKNORMAL_ROUTINE *NormalRoutine,
	IN OUT PVOID *NormalContext,
	IN OUT PVOID *SystemArgument1,
	IN OUT PVOID *SystemArgument2 )
{
	InApcCtxSwitches = (unsigned long) *((CHAR *) pkCurrThr + 0x58);
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; Hello from the kernel routine; init ctx sw = %d, in APC ctx sw = %d",
		KeGetCurrentThread(),
		InitCtxSwitches,
		InApcCtxSwitches);
	return;
}


//++
// Function:	
//
// Description:
//
// Arguments:
//
// Return value:
//
//--
static VOID ApcKernelRoutine2 (
	IN struct _KAPC *Apc,
	IN OUT PKNORMAL_ROUTINE *NormalRoutine,
	IN OUT PVOID *NormalContext,
	IN OUT PVOID *SystemArgument1,
	IN OUT PVOID *SystemArgument2 )
{
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; Hello from the kernel routine.",
		KeGetCurrentThread());
	return;
}


//++
// Function:	ApcKernelRoutineSIT
//
// Description:
//
// KernelRoutine for the spurious APC interrupt test.
//
// Arguments:
//
// Return value:
//
//--
static VOID ApcKernelRoutineSIT (
	IN struct _KAPC *Apc,
	IN OUT PKNORMAL_ROUTINE *NormalRoutine,
	IN OUT PVOID *NormalContext,
	IN OUT PVOID *SystemArgument1,
	IN OUT PVOID *SystemArgument2 )
{
	PDEVICE_EXTENSION pDevExt = *((PDEVICE_EXTENSION *) SystemArgument1);

	// Store a record tracking that the kernel routine
	// has been called
	//
	KERNEL_ROUTINE_RECORD KrRec;
	KrRec.RecordType = KernelRoutineRecord;
	KrRec.Thread.Thread = KeGetCurrentThread();
	KrRec.Thread.KernelApcPending = KernelApcPending(KrRec.Thread.Thread);
	KrRec.Thread.KApcListEmpty = KApcListEmpty(KrRec.Thread.Thread);
	KrRec.Thread.UserApcPending = UserApcPending(KrRec.Thread.Thread);
	KrRec.Thread.UApcListEmpty = UApcListEmpty(KrRec.Thread.Thread);

	__asm cli;

	NTSTATUS status = CBAllocateBlock(
		&BufCtl,
		sizeof KrRec);
	bBufferOverflow |= (status == CB_STATUS_WARN_DATA_LOST);
	PCHAR pData = (PCHAR) &KrRec;
	SIZE_T cbLen = sizeof KrRec;
	SIZE_T Offset = 0;

	CBStore(
		&BufCtl,
		&pData,
		&cbLen,
		&Offset,
		NULL);

	__asm sti;

	KeSetEvent(&pDevExt->Event, 0, FALSE);
	return;
}


//++
// Function:	ApcDisableTest
//
// Description:
//
//
// Arguments:
//
//	pDevObj - standard dispatch routine argument.
//
//	pIrp - standard dispatch routine argument.
//
// Return value:
//
//--
static NTSTATUS ApcDisableTest(
	PDEVICE_OBJECT		pDevObj, 
	PIRP				pIrp)
{
	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION) pDevObj->DeviceExtension;
	pkCurrThr = KeGetCurrentThread();
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\n\n\nAPCTEST - Thr: %#p; *** APC DISABLE TEST ***\n\n", 
		pkCurrThr);
	LARGE_INTEGER wait;
	wait.QuadPart = - 3 * 10 * 1000 * 1000; // 3 sec


	// Special kernel mode APC test
	//
	KeInitializeApc(
		&pDevExt->Apc,
		pkCurrThr,
		CurrentApcEnvironment,
		ApcKernelRoutine2,
		NULL,
		NULL,
		KernelMode,
		NULL);


	// Regular kernel mode APC test
	//
	//KeInitializeApc(
	//	&pDevExt->Apc,
	//	pkCurrThr,
	//	CurrentApcEnvironment,
	//	ApcKernelRoutine,
	//	NULL,
	//	KApcNormalRoutine,
	//	KernelMode,
	//	NULL);

	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; APC initialized: %#p", 
		pkCurrThr, &pDevExt->Apc);

	KEVENT Event;
	KeInitializeEvent(&Event, NotificationEvent, FALSE);

	// Get the address of the kernel APC list for the
	// current environment
	//
	LIST_ENTRY *pApcListHeads = (LIST_ENTRY *)((CHAR *) pkCurrThr + 0x38);
	LIST_ENTRY *pKApcList = &pApcListHeads[KernelMode];

	// Insert our APC at the head of the list
	//
	LIST_ENTRY *pApcListLink = (LIST_ENTRY *) ((CHAR *) &pDevExt->Apc + 0xc);

	__asm {
		cli
	}
	pKApcList->Flink->Blink = pApcListLink;
	pApcListLink->Flink = pKApcList->Flink;
	pApcListLink->Blink = pKApcList;
	pKApcList->Flink = pApcListLink;

	// Set KernelApcPending;
	//
	*((CHAR *) pkCurrThr + 0x38 + 0x15) = 1;

	// Set SpecialApcDisable
	//
	*((CHAR *) pkCurrThr + 0x82 ) = 1;


	__asm sti;

	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; APC inserted and flags set. Waiting...", 
		pkCurrThr);

	// Invoke a wait function to ensure thread dispatching occurs
	//
	KeWaitForSingleObject(&Event, Executive, KernelMode, TRUE, &wait);

	// Test KernelApcPending
	//
	volatile CHAR kApcPending = *((CHAR *) pkCurrThr + 0x38 + 0x15);
	if (kApcPending) {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - Thr: %#p; KernelApcPending STILL SET", 
			pkCurrThr);
	} else {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - Thr: %#p; KernelApcPending FOUND CLEAR", 
			pkCurrThr);
	}

	// Invoke KiDeliverApc
	//
	KIRQL PrevIrql;
	KeRaiseIrql((KIRQL) 1, &PrevIrql);
	KiDeliverApc(0, NULL, NULL);
	KeLowerIrql(PrevIrql);

	// Test KernelApcPending again
	//
	kApcPending = *((CHAR *) pkCurrThr + 0x38 + 0x15);
	if (kApcPending) {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - Thr: %#p; after KiDeliverAprc KernelApcPending STILL SET", 
			pkCurrThr);
	} else {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - Thr: %#p; after KiDeliverAprc KernelApcPending FOUND CLEAR", 
			pkCurrThr);
	}


	// Clear SpecialApcDisable
	//
	__asm cli;
	*((CHAR *) pkCurrThr + 0x82 ) = 0;
	__asm sti;

	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; SpecialApcDisable cleared. Waiting...", 
		pkCurrThr);

	// Invoke a wait function; no APC dispatching occurs, because
	// KernelApcPending has been cleared
	//
	KeWaitForSingleObject(&Event, Executive, KernelMode, TRUE, &wait);

	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; about to set KernelApcPending...", 
		pkCurrThr);


	// Set KernelApcPending;
	//
	__asm cli;
	*((CHAR *) pkCurrThr + 0x38 + 0x15) = 1;
	__asm sti;

	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; KernelApcPendins set. Waiting...", 
		pkCurrThr);

	// Invoke a wait function; we will see APC dispatching because of
	// SwapContext.
	//
	KeWaitForSingleObject(&Event, Executive, KernelMode, TRUE, &wait);


	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; after KeWaitForSingleObject()", 
		pkCurrThr);


	return STATUS_SUCCESS;
}


//++
// Function:	ApcRequester
//
// Description:
//
//
// Arguments:
//
//
// Return value:
//
//--
static VOID ApcRequester(
	 IN PVOID		StartContext)
{
	LARGE_INTEGER wait;
	wait.QuadPart = - 5 * 10 * 1000 * 1000; // 5 sec
	PRKTHREAD pkApcThr = KeGetCurrentThread();
	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION) StartContext;

	// Initialize a user mode APC targeted at the thread executing
	// the dispatch routine.
	//
	KeInitializeApc(
		&pDevExt->Apc,
		pkCurrThr,				// Static pointing at KTHREAD for the main thread
		CurrentApcEnvironment,
		UApcKernelRoutine,
		NULL,
		KApcNormalRoutine,		// We will not actually call this in user mode,
								// because UApcKernelRoutine will zero the pointer.
		UserMode,
		NULL);

	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - APC req thr: %#p; APC initialized: %#p", 
		pkApcThr, &pDevExt->Apc);

	// Wait before requesting the Apc
	//
	KEVENT Event;
	KeInitializeEvent(&Event, NotificationEvent, FALSE);
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - APC req thr: %#p; waiting before queuing the APC...", 
		pkApcThr);
	KeWaitForSingleObject(&Event, 
		Executive, 
		KernelMode, 
		FALSE, 
		&wait);

	// Request the APC
	//
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - APC req thr: %#p; Queuing an APC to thread %#p...", 
		pkApcThr,
		pkCurrThr);
	if (KeInsertQueueApc(
		&pDevExt->Apc,
		NULL,
		NULL,
		0)) {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - APC req thr: %#p; APC queued: %#p", 
			pkApcThr, &pDevExt->Apc);
	} else {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - APC req thr: %#p; Failed to queue APC",
			pkApcThr);
	}



	PsTerminateSystemThread(STATUS_SUCCESS);
}


//++
// Function:	ApcSpuriousIntTest
//
// Description:
//
//
// Arguments:
//
//	pDevObj - standard dispatch routine argument.
//
//	pIrp - standard dispatch routine argument.
//
// Return value:
//
//--
static NTSTATUS ApcSpuriousIntTest(
	PDEVICE_OBJECT		pDevObj, 
	PIRP				pIrp)
{
	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION) pDevObj->DeviceExtension;
	pkCurrThr = KeGetCurrentThread();
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\n\n\nAPCTEST - Thr: %#p; *** APC SPURIOUS INTERRUPT TEST ***\n\n", 
		pkCurrThr);
	LARGE_INTEGER wait;
	wait.QuadPart = - 10 * 10 * 1000 * 1000; // 10 sec
	KeInitializeEvent(&pDevExt->Event, NotificationEvent, FALSE);


	// Initialize the circular buffer;
	//
	CBInitBufferCtlVar(
		&BufCtl, 
		BufferStore,
		sizeof BufferStore);
	bBufferOverflow = FALSE;



	// Initialize a special kernel mode APC
	//
	KeInitializeApc(
		&pDevExt->Apc,
		pkCurrThr,
		CurrentApcEnvironment,
		ApcKernelRoutineSIT,
		NULL,
		NULL, //NormalRoutine
		NULL,
		pDevExt);
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; APC initialized: %#p", 
		pkCurrThr, &pDevExt->Apc);

	// Get a pointer to the _KPCR
	//
	PCHAR pKpcr = GetKpcr();


	// Set the IRQL to dispatch to prevent dispatching
	//
	KIRQL prevIrql;
	KeRaiseIrql((KIRQL) 2, &prevIrql);

	// Queue the APC while IRQL = DISPATCH: the APC int cannot
	// be serviced until the IRQL drops to PASSIVE
	//
	KeInsertQueueApc(
		&pDevExt->Apc,
		pDevExt,
		NULL,
		0);

	// Spin in a loop until a DISPATCH int is pending or
	// a timeout expires.
	//
	UINT32 bNoDispInt = 0;
	LARGE_INTEGER FinalTickCount, TickCount;
	volatile CHAR DispIntFlag;
	INT WaitCycles = 0;
	KeQueryTickCount(&FinalTickCount);
	FinalTickCount.QuadPart += 50000;
	for (WaitCycles = 1;;WaitCycles++) {
		DispIntFlag = *(pKpcr + 0x95 + 0x2);	// This byte is set to 1 when a
			// DISPATCH interrupt is pending
			//
		if (DispIntFlag) break;
		KeQueryTickCount(&TickCount);
		if (TickCount.QuadPart > FinalTickCount.QuadPart) {
			bNoDispInt = 1;
			break;
		}
	}

	// If we failed to wait for the DISPATCH int,
	// leave.
	//
	if (bNoDispInt) {
		KeLowerIrql(prevIrql);
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - Thr: %#p; Failed to wait for DISPATCH interrupt",
			pkCurrThr);
		return STATUS_SUCCESS;
	}


	// Preset the trace switch to on. The hook of nt!KiDeliverAPC
	// will turn it off when it detects an APC for this thread.
	//
	bHookTraceOn = TRUE;


	__asm cli;

	// Hook nt!SwapContext and nt!KiDeliverApc
	//
	SetHooks();

	__asm sti;

	// Now we have both the APC and the DISPATCH int pending.
	// Lower the IRQL. The hooks trace the context switches
	// and the spurious APC int
	//
	KeLowerIrql(prevIrql);

	// Store a trace record tracing the return
	// from KeLowerIrql()
	//
	MESSAGE_RECORD MsgRecord;
	MsgRecord.RecordType = MessageRecord;
	MsgRecord.MessageIndex = 0;

	__asm cli;

	NTSTATUS status = CBAllocateBlock(
		&BufCtl,
		sizeof MsgRecord);
	bBufferOverflow |= (status == CB_STATUS_WARN_DATA_LOST);
	PCHAR pData = (PCHAR) &MsgRecord;
	SIZE_T cbLen = sizeof MsgRecord;
	SIZE_T Offset = 0;
	CBStore(
		&BufCtl,
		&pData,
		&cbLen,
		&Offset,
		NULL);

	__asm sti;



	// Wait for the kernel routine
	//
	status = KeWaitForSingleObject(
		&pDevExt->Event,
		Executive,
		KernelMode,
		FALSE,
		&wait);


	__asm cli;

	RemoveHooks();

	__asm sti;

	// Dump the trace buffer content to the debugger console
	//
	DumpTrace();
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; WaitCycles = %d",
		pkCurrThr,
		WaitCycles);
	return STATUS_SUCCESS;
}


//++
// Function:	ApcUserModeTest
//
// Description:
//
//
// Arguments:
//
//	pDevObj - standard dispatch routine argument.
//
//	pIrp - standard dispatch routine argument.
//
// Return value:
//
//--
static NTSTATUS ApcUserModeTest(
	PDEVICE_OBJECT		pDevObj, 
	PIRP				pIrp)
{
	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION) pDevObj->DeviceExtension;
	pkCurrThr = KeGetCurrentThread();
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\n\n\nAPCTEST - Thr: %#p; *** APC USER MODE TEST ***\n\n", 
		pkCurrThr);
	LARGE_INTEGER wait;
	wait.QuadPart = - 10 * 10 * 1000 * 1000; // 10 sec
	NTSTATUS status;

	KeInitializeEvent(&pDevExt->Event, NotificationEvent, FALSE);


	// Create a separate thread which will queue the APC
	//
	HANDLE hThread;
	OBJECT_ATTRIBUTES ObjectAttributes;
	InitializeObjectAttributes(
		&ObjectAttributes, 
		NULL, 
		OBJ_KERNEL_HANDLE, 
		NULL, 
		NULL);
	status = PsCreateSystemThread(
		&hThread,
		THREAD_ALL_ACCESS,
		&ObjectAttributes,
		NULL,
		NULL,
		ApcRequester,
		pDevExt);
	if (!NT_SUCCESS(status)) {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - Thr: %#p; PsCreateSystemThread failed with status: %#x", 
			pkCurrThr, status);
			return status;
	}
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; entering an alertable user-mode wait...", 
		pkCurrThr);

	NTSTATUS WaitStatus = KeWaitForSingleObject(&pDevExt->Event, 
		Executive, 
		UserMode, 
		TRUE, 
		&wait);

	if (WaitStatus == STATUS_USER_APC) {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - Thr: %#p; KeWaitForSingleObject returned STATUS_USER_APC", 
			pkCurrThr);
	} else {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - Thr: %#p; KeWaitForSingleObject returned %x", 
			pkCurrThr,
			WaitStatus);
	}

	// Try to wait some more in user mode
	//
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; entering again an alertable user-mode wait...", 
		pkCurrThr);

	WaitStatus = KeWaitForSingleObject(&pDevExt->Event, 
		Executive, 
		UserMode, 
		TRUE, 
		&wait);

	if (WaitStatus == STATUS_USER_APC) {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - Thr: %#p; KeWaitForSingleObject returned STATUS_USER_APC", 
			pkCurrThr);
	} else {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - Thr: %#p; KeWaitForSingleObject returned %x", 
			pkCurrThr,
			WaitStatus);
	}

	// Try to wait in non-alertable user mode
	//
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; entering a non-alertable user-mode wait...", 
		pkCurrThr);

	WaitStatus = KeWaitForSingleObject(&pDevExt->Event, 
		Executive, 
		UserMode, 
		FALSE, 
		&wait);

	if (WaitStatus == STATUS_USER_APC) {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - Thr: %#p; KeWaitForSingleObject returned STATUS_USER_APC", 
			pkCurrThr);
	} else {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - Thr: %#p; KeWaitForSingleObject returned %x", 
			pkCurrThr,
			WaitStatus);
	}

	// Try to wait in non-alertable kernel mode
	//
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; entering a non-alertable kernel-mode wait...", 
		pkCurrThr);

	WaitStatus = KeWaitForSingleObject(&pDevExt->Event, 
		Executive, 
		KernelMode, 
		FALSE, 
		&wait);

	if (WaitStatus == STATUS_TIMEOUT) {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - Thr: %#p; KeWaitForSingleObject returned STATUS_TIMEOUT", 
			pkCurrThr);
	} else {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - Thr: %#p; KeWaitForSingleObject returned %x", 
			pkCurrThr,
			WaitStatus);
	}

	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; returning to user mode", 
		pkCurrThr);


	return STATUS_SUCCESS;
}


//++
// Function:	ApcUserModeTest2
//
// Description:
//
//
// Arguments:
//
//	pDevObj - standard dispatch routine argument.
//
//	pIrp - standard dispatch routine argument.
//
// Return value:
//
//--
static NTSTATUS ApcUserModeTest2(
	PDEVICE_OBJECT		pDevObj, 
	PIRP				pIrp)
{
	PDEVICE_EXTENSION pDevExt = (PDEVICE_EXTENSION) pDevObj->DeviceExtension;
	pkCurrThr = KeGetCurrentThread();
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\n\n\nAPCTEST - Thr: %#p; *** APC USER MODE TEST #2 ***\n\n", 
		pkCurrThr);
	LARGE_INTEGER wait;
	wait.QuadPart = - 10 * 10 * 1000 * 1000; // 10 sec


	// Initialize a user mode APC targeted at this thread
	//
	KeInitializeApc(
		&pDevExt->Apc,
		pkCurrThr,				
		CurrentApcEnvironment,
		UApcKernelRoutine,
		NULL,
		KApcNormalRoutine,		// We will not actually call this in user mode,
								// because UApcKernelRoutine will zero the pointer.
		UserMode,
		NULL);

	// Check UserApcPending
	//
	volatile INT UserApcPending = *(((UCHAR *) pkCurrThr) + 0x38 + 0x16);
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - thr: %#p; UserApcPending BEFORE requesting the APC: %i", 
		pkCurrThr,
		UserApcPending);

	// Request the APC
	//
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - thr: %#p; Queuing the APC...", 
		pkCurrThr);
	if (KeInsertQueueApc(
		&pDevExt->Apc,
		NULL,
		NULL,
		0)) {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - thr: %#p; APC queued: %#p", 
			pkCurrThr, &pDevExt->Apc);
	} else {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - thr: %#p; Failed to queue APC",
			pkCurrThr);
	}

	// Check UserApcPending again
	//
	UserApcPending = *(((UCHAR *) pkCurrThr) + 0x38 + 0x16);
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - thr: %#p; UserApcPending AFTER requesting the APC: %i", 
		pkCurrThr,
		UserApcPending);

	// Try to enter an user-mode alertalbe wait
	//
	KeInitializeEvent(&pDevExt->Event, NotificationEvent, FALSE);
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; entering an alertable user-mode wait...", 
		pkCurrThr);

	NTSTATUS WaitStatus = KeWaitForSingleObject(&pDevExt->Event, 
		Executive, 
		UserMode, 
		TRUE, 
		&wait);

	if (WaitStatus == STATUS_USER_APC) {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - Thr: %#p; KeWaitForSingleObject returned STATUS_USER_APC", 
			pkCurrThr);
	} else {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - Thr: %#p; KeWaitForSingleObject returned %x", 
			pkCurrThr,
			WaitStatus);
	}

	// Check UserApcPending again
	//
	UserApcPending = *(((UCHAR *) pkCurrThr) + 0x38 + 0x16);
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - thr: %#p; UserApcPending AFTER KeWaitForSingleObject: %i", 
		pkCurrThr,
		UserApcPending);

	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; returning to user mode", 
		pkCurrThr);


	return STATUS_SUCCESS;
}


//++
// Function:	CreateDevice
//
// Description:
//		Adds a new device.
//
//		Initializes the device extension.
//
// Arguments:
//		pDriverObject - Passed from I/O Manager
//
//		ppDevExt - pointer to a pointer which will receive the
//			address of the device extension returned by the
//			I/O manager.
//
// Return value:
//		NTSTATUS. Possible failure codes:
//
//			Any failure code returned by IoCreateDevice().
//
//			Any failure code returned by IoCreateSymbolicLink().
//
//--
static NTSTATUS CreateDevice (
		IN PDRIVER_OBJECT	pDriverObject,
		OUT PDEVICE_EXTENSION*	ppDevExt	) 
{

	NTSTATUS status;
	PDEVICE_OBJECT pDevObj;
	PDEVICE_EXTENSION pDevExt;
	UNICODE_STRING uRecName, uSymName;

	RtlInitUnicodeString(&uRecName,
        DRV_INT_NAME);

	// Now create the device
	status =
		IoCreateDevice( pDriverObject,
						sizeof(DEVICE_EXTENSION),
						&uRecName,
						FILE_DEVICE_UNKNOWN,
						0, FALSE,
						&pDevObj );
	if (!NT_SUCCESS(status))
		return status;

	// Initialize the Device Extension
	pDevExt = (PDEVICE_EXTENSION)pDevObj->DeviceExtension;

	pDevExt->pDevice = pDevObj;	// back pointer
	*ppDevExt = pDevExt;

	// Form the symbolic link name
	RtlInitUnicodeString(&uSymName,
        DRV_SYM_LINK);
	status = 
		IoCreateSymbolicLink( &uSymName,
							  &uRecName );
	if (!NT_SUCCESS(status)) {

		// if it fails now, must delete Device object
		IoDeleteDevice( pDevObj );
		return status;
	}

	// Made it
	return STATUS_SUCCESS;
}


//++
// Function:	DeleteDevices
//
// Description:
//		Deletes all the devices created by this driver.
//
// Arguments:
//		pDriverObject - pointer to driver object
//
// Return value:
//		NTSTATUS - success or failure code
//--

NTSTATUS DeleteDevices(IN PDRIVER_OBJECT pDriverObject) 
{

	PDEVICE_OBJECT	pNextObj;
	UNICODE_STRING  uSymName;

	pNextObj = pDriverObject->DeviceObject;
	if (pNextObj != NULL) {

		// Form the symbolic link which must be deleted
        RtlInitUnicodeString(&uSymName, DRV_SYM_LINK);
		IoDeleteSymbolicLink(&uSymName);

		// Delete the device.
		IoDeleteDevice( pNextObj );
	}
	return STATUS_SUCCESS;

}



//++
// Function:	DispatchClose
//
// Description:
//		Handles call from Win32 CreateHandle request
//		For this driver, does nothing
//
// Arguments:
//		pDevObj - Passed from I/O Manager
//		pIrp - Passed from I/O Manager
//
// Return value:
//		NTSTATUS - always returns STATUS_SUCCESS;
//--

NTSTATUS DispatchClose (
		IN PDEVICE_OBJECT	pDevObj,
		IN PIRP				pIrp			) {
 
	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;	// no bytes xfered
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );



	return STATUS_SUCCESS;
}


//++
// Function:	DispatchCreate
//
// Description:
//		Handles call from Win32 CreateFile request
//		For this driver, does nothing
//
// Arguments:
//		pDevObj - Passed from I/O Manager
//		pIrp - Passed from I/O Manager
//
// Return value:
//		NTSTATUS - always returns STATUS_SUCCESS;
//--


NTSTATUS DispatchCreate (
		IN PDEVICE_OBJECT	pDevObj,
		IN PIRP				pIrp			) {

	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;	// no bytes xfered
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );


	return STATUS_SUCCESS;
}

//++
// Function:	DispatchDioc
//
// Description:
//		Handles call from Win32 DeviceIoControl request.
//
// Arguments:
//		pDevObj - Passed from I/O Manager
//		pIrp - Passed from I/O Manager
//
// Return value:
//		NTSTATUS - success or failure code
//--


NTSTATUS DispatchDioc (
		IN PDEVICE_OBJECT	pDevObj,
		IN PIRP				pIrp			) {



	NTSTATUS status = STATUS_SUCCESS;

	PIO_STACK_LOCATION pIrpStack = IoGetCurrentIrpStackLocation(pIrp);
	ULONG ctlCode = pIrpStack->Parameters.DeviceIoControl.IoControlCode;

	pIrp->IoStatus.Information = 0;	// bytes xfered
	
	switch (ctlCode) {

        // Insert driver - specific control codes here.

	case IOCTL_APCTEST_DISABLE_TEST:
		status = ApcDisableTest(pDevObj, pIrp);
		break;
	case IOCTL_APCTEST_USER_MODE_TEST:
		status = ApcUserModeTest(pDevObj, pIrp);
		break;
	case IOCTL_APCTEST_USER_MODE_TEST2:
		status = ApcUserModeTest2(pDevObj, pIrp);
		break;
	case IOCTL_APCTEST_SPURIOUS_INT_TEST:
		status = ApcSpuriousIntTest(pDevObj, pIrp);
		break;
	default:
		status = STATUS_INVALID_DEVICE_REQUEST;
		break;
	}

	// If the request isn't complete, return STATUS_PENDING.
	if (status == STATUS_PENDING)
		return status;

    // Note: IoMarkIrpPending must have already been called
    // by the function which returned STATUS_PENDING.

	// Otherwise, complete the IRP.
	pIrp->IoStatus.Status = status;
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );
	return status;

}


//++
// Function:	DispatchRead
//
// Description:
//		Handles call for IRP_MJ_READ
//
// Arguments:
//		pDevObj - Passed from I/O Manager
//		pIrp - Passed from I/O Manager
//
// Return value:
//		NTSTATUS - success or failure code
//--

NTSTATUS DispatchRead(
	  IN PDEVICE_OBJECT	pDevObj,
	  IN PIRP			pIrp )
{

	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"APCTEST - DispatchRead() called.");

	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;
	IoCompleteRequest(pIrp, IO_NO_INCREMENT);


	return STATUS_SUCCESS;
}


//++
// Function:	DispatchWrite
//
// Description:
//		Handles call for IRP_MJ_WRITE
//
// Arguments:
//		pDevObj - Passed from I/O Manager
//		pIrp - Passed from I/O Manager
//
// Return value:
//		NTSTATUS - success or failure code
//--

NTSTATUS DispatchWrite(
	  IN PDEVICE_OBJECT	pDevObj,
	  IN PIRP			pIrp )
{

	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"APCTEST - DispatchWrite() called.");

	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;
	IoCompleteRequest(pIrp, IO_NO_INCREMENT);

	return STATUS_SUCCESS;
}


//++
// Function:	
//
// Description:
//
// Arguments:
//
// Return value:
//
//--
static VOID DumpTrace ()
{

	if (bBufferOverflow) {
		DbgPrintEx(
			DPFLTR_DEFAULT_ID,
			DPFLTR_ERROR_LEVEL,
			"\nAPCTEST - Buffer overflow, the trace is not complete");
	}


	TRACE_RECORD TrcRec;


	NTSTATUS status;
	
	for (status = STATUS_SUCCESS; status == STATUS_SUCCESS;) {
		SIZE_T Offset = 0;
		PCHAR pDest = (PCHAR) &TrcRec;
		SIZE_T cbBufSize = sizeof TrcRec;
		status =  CBXtractFromStart(
			&BufCtl,
			&Offset,
			&pDest,
			&cbBufSize,
			NULL,
			NULL);
		if (!NT_SUCCESS(status)) break;
		PrintTraceRecord(&TrcRec);
		CBFreeOldestDataBlock(&BufCtl);
	}

}


//++
// Function:	GetKpcr
//
// Description:
//
// Returns the address of the _KPCR for the executing processor.
//
// It is based on the fact that fs:[0] is equal to the linear address of 
// the _KPCR and that, at +0x1c, _KPCR.SelfPcr stores the linear address 
// itself.
//
// Arguments:
//
// Return value:
//
//--
static PCHAR GetKpcr()
{
	PCHAR pKpcr;
	__asm {
		push eax
		mov eax, dword ptr fs:[1ch]
		mov pKpcr, eax
		pop eax
	}
	return pKpcr;
}


//++
// Function:	GetIrql
//
// Returns the current IRQL, which is stored inside the
// _KPCR for the executing processor.
//
// Description:
//
// Arguments:
//
// Return value:
//
//--
static KIRQL GetIrql()
{
	return *(GetKpcr() + IRQL_OFFSET);

}


//++
// Function:	KApcListEmpty
//
// Description:
//
// Returns TRUE if the kernel APC list is empty.
//
// Arguments:
//
// Return value:
//
//--
static BOOLEAN KApcListEmpty(
	PKTHREAD		Thread)
{
	// Compute the address of the K APClist head
	//
	PCHAR pKApcListHead = (PCHAR) Thread + APC_STATE_OFFSET;

	PCHAR pNext = *((PCHAR *) pKApcListHead);
	return (pNext == pKApcListHead);
}


//++
// Function:	
//
// Description:
//
// Arguments:
//
// Return value:
//
//--
VOID KApcNormalRoutine(
	   PVOID		NormalContext,
	   PVOID		SystemArgument1,
	   PVOID		SystemArgument2)
{
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; Hello from the normal routine. Irql = %d",
		KeGetCurrentThread(),
		KeGetCurrentIrql());
	return;
}


//++
// Function:	KernelApcPending
//
// Description:
//
// Returns TRUE if KernelApcPending is set.
//
// Arguments:
//
// Return value:
//
//--
static BOOLEAN KernelApcPending(
		PKTHREAD		Thread)
{
	
	return *((PCHAR) Thread + APC_STATE_OFFSET + KERNEL_APC_PENDING_OFFSET);
}


//++
// Function:	KiDeliverApcTrace
//
// Description:
//
//	Called by the hook of nt!KiDeliverApc.
//
// Arguments:
//
//	CurrThr - Current Thread.
//
//	pCaller - Return address into the calling function.
//
// Return value:
//
//--
extern "C" VOID __cdecl KiDeliverApcTrace(
  PCHAR		TrapFrame,
  PVOID		Reserved,
  INT		PreviousMode,
  PCHAR		pCaller,
  PKTHREAD	CurrThr)
{
	if (!bHookTraceOn) return;

	KI_DELIVER_APC_RECORD KdapRecord;
	KdapRecord.RecordType = KiDeliverApcRecord;
	KdapRecord.Irql = GetIrql();
	KdapRecord.Caller = pCaller;
	KdapRecord.TrapFrame = TrapFrame;
	KdapRecord.Reserved = Reserved;
	KdapRecord.PreviousMode = PreviousMode;
	KdapRecord.Thread.Thread = CurrThr;
	KdapRecord.Thread.KernelApcPending = KernelApcPending(CurrThr);
	KdapRecord.Thread.KApcListEmpty = KApcListEmpty(CurrThr);
	KdapRecord.Thread.UserApcPending = UserApcPending(CurrThr);
	KdapRecord.Thread.UApcListEmpty = UApcListEmpty(CurrThr);

	NTSTATUS status = CBAllocateBlock(
		&BufCtl,
		sizeof KdapRecord);
	bBufferOverflow |= (status == CB_STATUS_WARN_DATA_LOST);
	PCHAR pData = (PCHAR) &KdapRecord;
	SIZE_T cbLen = sizeof KdapRecord;
	SIZE_T Offset = 0;
	CBStore(
		&BufCtl,
		&pData,
		&cbLen,
		&Offset,
		NULL);

	// If we get called into our thread, stop the trace
	//
	bHookTraceOn = !(pkCurrThr == CurrThr);

	return;
}


//++
// Function:	
//
// Description:
//
// Arguments:
//
// Return value:
//
//--
static VOID PrintTraceRecord(
	PTRACE_RECORD	pTrcRec)
{
	switch(pTrcRec->RecordType) {
		case SwapContextRecord:
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\n\nAPCTEST - SWAP CONTEXT trace\n\n");
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		Current IRQL:                    %#hx",
				pTrcRec->SwapContextRecord.Irql);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		Current thread:                  %#p",
				pTrcRec->SwapContextRecord.CurrThr.Thread);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		Current thread K APC pending:    %hx",
				pTrcRec->SwapContextRecord.CurrThr.KernelApcPending);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		Current thread K APC list empty: %hx",
				pTrcRec->SwapContextRecord.CurrThr.KApcListEmpty);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		Current thread U APC pending:    %hx",
				pTrcRec->SwapContextRecord.CurrThr.UserApcPending);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		Current thread U APC list empty: %hx",
				pTrcRec->SwapContextRecord.CurrThr.UApcListEmpty);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\n"
				"\nAPCTEST -		New thread:                      %#p",
				pTrcRec->SwapContextRecord.NewThr.Thread);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		New thread K APC pending:        %hx",
				pTrcRec->SwapContextRecord.NewThr.KernelApcPending);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		New thread K APC list empty:     %hx",
				pTrcRec->SwapContextRecord.NewThr.KApcListEmpty);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		New thread U APC pending:        %hx",
				pTrcRec->SwapContextRecord.NewThr.UserApcPending);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		New thread U APC list empty:     %hx",
				pTrcRec->SwapContextRecord.NewThr.UApcListEmpty);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\n"
				"\nAPCTEST -		APC INT:                         %hx",
				pTrcRec->SwapContextRecord.ApcInt);
			break;
		case KiDeliverApcRecord:
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\n\nAPCTEST - DELIVER APC trace\n\n");
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		Current IRQL:                    %#hx",
				pTrcRec->DeliverApcRecord.Irql);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		Caller address:                  %#p",
				pTrcRec->DeliverApcRecord.Caller);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		Trap frame:                      %#p",
				pTrcRec->DeliverApcRecord.TrapFrame);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		Reserved:                        %#p",
				pTrcRec->DeliverApcRecord.Reserved);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		PreviousMode:                    %d",
				pTrcRec->DeliverApcRecord.PreviousMode);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\n"
				"\nAPCTEST -		Thread:                          %#p",
				pTrcRec->DeliverApcRecord.Thread.Thread);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		Thread K APC pending:            %hx",
				pTrcRec->DeliverApcRecord.Thread.KernelApcPending);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		Thread K APC list empty:         %hx",
				pTrcRec->DeliverApcRecord.Thread.KApcListEmpty);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		Thread U APC pending:            %hx",
				pTrcRec->DeliverApcRecord.Thread.UserApcPending);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		Thread U APC list empty:         %hx",
				pTrcRec->DeliverApcRecord.Thread.UApcListEmpty);
			break;
		case KernelRoutineRecord:
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\n\nAPCTEST - KERNEL ROUTINE trace\n\n");
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\n"
				"\nAPCTEST -		Thread:                          %#p",
				pTrcRec->KernelRoutineRecord.Thread.Thread);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		Thread K APC pending:            %hx",
				pTrcRec->KernelRoutineRecord.Thread.KernelApcPending);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		Thread K APC list empty:         %hx",
				pTrcRec->KernelRoutineRecord.Thread.KApcListEmpty);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		Thread U APC pending:            %hx",
				pTrcRec->KernelRoutineRecord.Thread.UserApcPending);
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST -		Thread U APC list empty:         %hx",
				pTrcRec->KernelRoutineRecord.Thread.UApcListEmpty);
			break;
		case MessageRecord:
			DbgPrintEx(
				DPFLTR_DEFAULT_ID,
				DPFLTR_ERROR_LEVEL,
				"\nAPCTEST - TRACE MESSAGE: %s",
				szTraceMessages[pTrcRec->MessageRecord.MessageIndex]);
			break;

	}

}



//++
// Function:	
//
// Description:
//
// Arguments:
//
// Return value:
//
//--
static VOID RemoveHooks()
{
	UnhookSWCN();
	UnhookKDAP();
}


//++
// Function:	
//
// Description:
//
// Arguments:
//
// Return value:
//
//--
static VOID SetHooks()
{
// Offset of KeWaitForSingleObject()
//
#define OFF_KEWAITFORSINGLEOBJECT 0x547c9

// Offset of SwapContext
//
#define OFF_SWAPCONTEXT			0x690a0

// Offset of KiDeliverApc
//
#define OFF_KIDELIVERAPC		0x7a08c

	// Compute the hooked addresses.
	//
	// To compute the run-time addresses, use the address of
	// an exported function (e. g. KeWaitForSingleObject), then subtract
	// its offset (obtained from the debugger) from the address. This
	// gives us the load address of ntoskrnl.exe. Then we add to it
	// the offset of the desired symbol, also obtained from the debugger.
	//
	// This is necessary, b/c SwapContext and KiDeliverApc are not exported.
	//
	lpNtOsKrnl = (PCHAR) &KeWaitForSingleObject - OFF_KEWAITFORSINGLEOBJECT;
	lpSwapContext = lpNtOsKrnl + OFF_SWAPCONTEXT;
	lpKiDeliverApc = lpNtOsKrnl + OFF_KIDELIVERAPC;

	// HookSc hooks SwapContext
	//
	SetSWCNHook();
	SetKDAPHook();
}


//++
// Function:	SwapContextTrace
//
// Description:
//
// Called by the hook of nt!SwapContext.
//
// Arguments:
//
//	CurrThr - thread about to be suspended
//
//	NewThr - thread about to be resumed
//
// Return value:
//
//--
extern "C" VOID __cdecl SwapContextTrace(
  PKTHREAD		CurrThr,
  PKTHREAD		NewThr)
{
	if (!bHookTraceOn) return;

	// Get a pointer to the _KPCR
	//
	PCHAR pKpcr = GetKpcr();

	SWAP_CONTEXT_RECORD SwctRecord;
	SwctRecord.RecordType = SwapContextRecord;
	SwctRecord.Irql = GetIrql();
	SwctRecord.CurrThr.Thread = CurrThr;
	SwctRecord.CurrThr.KernelApcPending = KernelApcPending(CurrThr);
	SwctRecord.CurrThr.KApcListEmpty = KApcListEmpty(CurrThr);
	SwctRecord.CurrThr.UserApcPending = UserApcPending(CurrThr);
	SwctRecord.CurrThr.UApcListEmpty = UApcListEmpty(CurrThr);
	SwctRecord.NewThr.Thread = NewThr;
	SwctRecord.NewThr.KernelApcPending = KernelApcPending(NewThr);
	SwctRecord.NewThr.KApcListEmpty = KApcListEmpty(NewThr);
	SwctRecord.NewThr.UserApcPending = UserApcPending(CurrThr);
	SwctRecord.NewThr.UApcListEmpty = UApcListEmpty(CurrThr);

	// Store in the recrd the status of the ApcInt flag: true
	// if int pending.
	//
	SwctRecord.ApcInt = ApcInt(pKpcr);
	NTSTATUS status = CBAllocateBlock(
		&BufCtl,
		sizeof SwctRecord);
	bBufferOverflow |= (status == CB_STATUS_WARN_DATA_LOST);
	PCHAR pData = (PCHAR) &SwctRecord;
	SIZE_T cbLen = sizeof SwctRecord;
	SIZE_T Offset = 0;
	CBStore(
		&BufCtl,
		&pData,
		&cbLen,
		&Offset,
		NULL);

}


//++
// Function:	
//
// Description:
//
// Arguments:
//
// Return value:
//
//--
static VOID UApcKernelRoutine (
	IN struct _KAPC *Apc,
	IN OUT PKNORMAL_ROUTINE *NormalRoutine,
	IN OUT PVOID *NormalContext,
	IN OUT PVOID *SystemArgument1,
	IN OUT PVOID *SystemArgument2 )
{
	DbgPrintEx(
		DPFLTR_DEFAULT_ID,
		DPFLTR_ERROR_LEVEL,
		"\nAPCTEST - Thr: %#p; Hello from the USER APC kernel routine",
		KeGetCurrentThread());

	// Zero the NormalRoutine pointer to avoid its invocation
	//
	*NormalRoutine = 0;
	return;
}


//++
// Function:	UApcListEmpty
//
// Description:
//
// Returns TRUE if the user APC list is empty.
//
// Arguments:
//
// Return value:
//
//--
static BOOLEAN UApcListEmpty(
	PKTHREAD	Thread)
{
	// Compute the address of the U APClist head
	//
	PCHAR pUApcListHead = (PCHAR) Thread + APC_STATE_OFFSET + 8;
	PCHAR pNext = *((PCHAR *) pUApcListHead);
	return (pNext == pUApcListHead);
}


//++
// Function:	UserApcPending
//
// Description:
//
// Returns TRUE if UserApcPending ise set.
//
// Arguments:
//
// Return value:
//
//--
static BOOLEAN UserApcPending(
	PKTHREAD	Thread)
{
	return *((PCHAR) Thread + APC_STATE_OFFSET + USER_APC_PENDING_OFFSET);
}